// d3dhack.cpp : Defines the entry point for the DLL application.
//
#pragma warning(disable:4800)
#pragma warning(disable:4530)

//#define INITGUID
#include <stdio.h>

#include <map>
#include <list>
using namespace std;

#define CINTERFACE
#include <d3d8.h>
#define DIRECTINPUT_VERSION 0x0800
#include <dinput.h>
#undef CINTERFACE

#include <d3dx8.h>

#include "apihijack.h"

#include <shlwapi.h>

extern HINSTANCE hDLL;

struct STextureData
{	
	// Special rendering parameters.
	DWORD Style;
	D3DCOLOR Color1, Color2;
	float Rate;

	// Texture parameters.
	UINT Width, Height;
	DWORD Format;
	DWORD Checksum;

	// Data passed from 'Lock' to 'Unlock'.
	UINT Level;
	D3DLOCKED_RECT LockedRect;
	RECT Rect;
};

map<DWORD,STextureData> TextureBank;

list<DWORD> TextureList;
list<DWORD>::iterator CurTexture;

bool SaveTextures = false;

void ReadINIFile()
{
	OutputDebugString( "Reading INI file...\n" );
	TextureBank.clear();
	TextureList.clear();

	// Read INI file.
	char Path[MAX_PATH];
	GetModuleFileName( hDLL, Path, sizeof(Path) );
	PathRemoveFileSpec( Path );
	strcat( Path, "\\pshack.ini" );

	SaveTextures = GetPrivateProfileInt( "Options", "SaveTextures", 0, Path );

	static char Sections[2*1024];
	GetPrivateProfileSectionNames( Sections, sizeof(Sections), Path );

	for ( char* Section = Sections; *Section; Section += strlen(Section)+1 )
	{
		static char String[2*1024];
		
		STextureData Data;

		Data.Style = GetPrivateProfileInt( Section, "Style", 1, Path );

		GetPrivateProfileString( Section, "Color1", "0xFFFFFFFF", String, sizeof(String), Path );
		Data.Color1 = strtoul( String, NULL, 0 );

		GetPrivateProfileString( Section, "Color2", "0xFFFFFFFF", String, sizeof(String), Path );
		Data.Color2 = strtoul( String, NULL, 0 );

		GetPrivateProfileString( Section, "Rate", "0.0", String, sizeof(String), Path );
		Data.Rate = (float)strtod( String, NULL );

		GetPrivateProfileString( Section, "Checksum", "0", String, sizeof(String), Path );
		char* p = String;
		Data.Checksum = strtoul( p, &p, 16 );
		while ( Data.Checksum )
		{
			map<DWORD,STextureData>::iterator it = TextureBank.find( Data.Checksum );
			if ( it == TextureBank.end() )
			{
//				char Work[1024];
//				_snprintf( Work, sizeof(Work), "Added texture record. Section: %s Checksum: %x\n", Section, Data.Checksum);
//				OutputDebugString( Work );

				TextureBank.insert( pair<DWORD,STextureData>( Data.Checksum, Data ) );
				Data.Checksum = strtoul( p, &p, 16 );
			}
			else
			{
				char Work[1024];
				_snprintf( Work, sizeof(Work), "Duplicate texture record! Checksum: %x\n" );
				OutputDebugString( Work );
			}
		} 
	}

	TextureList.push_back( 0 );
	CurTexture = TextureList.begin();
}

// IDirect3DTexture8 Interface
IDirect3DTexture8Vtbl HookedIDirect3DTexture8Vtbl;
IDirect3DTexture8Vtbl OldIDirect3DTexture8Vtbl;

// {B82E3500-376B-42aa-8B53-10EB95905A4B}
static const GUID STextureDataGUID = 
{ 0xb82e3500, 0x376b, 0x42aa, { 0x8b, 0x53, 0x10, 0xeb, 0x95, 0x90, 0x5a, 0x4b } };

HRESULT __stdcall IDirect3DTexture8Hook_LockRect( 
	IDirect3DTexture8* _this, UINT Level, D3DLOCKED_RECT* pLockedRect, CONST RECT* pRect, DWORD Flags )
{
	HRESULT ret = OldIDirect3DTexture8Vtbl.LockRect( _this, Level, pLockedRect, pRect, Flags );

	STextureData* Data;
	DWORD Size = sizeof(Data);
	HRESULT ok = OldIDirect3DTexture8Vtbl.GetPrivateData( _this, STextureDataGUID, &Data, &Size );
	if ( ok == D3D_OK )
	{
		Data->Level = Level;
		Data->LockedRect = *pLockedRect;

		if ( pRect )
			Data->Rect = *pRect;
		else
		{
			Data->Rect.left = Data->Rect.top = 0;
			Data->Rect.right = Data->Width-1;
			Data->Rect.bottom = Data->Height-1;
		}
	}

	return ret;
}

HRESULT __stdcall IDirect3DTexture8Hook_UnlockRect( IDirect3DTexture8* _this, UINT Level )
{
	bool Save = false;

	STextureData* Data;
	DWORD Size = sizeof(Data);
	HRESULT ok = OldIDirect3DTexture8Vtbl.GetPrivateData( _this, STextureDataGUID, &Data, &Size );

	if ( ok == D3D_OK && Data->Level == 0 )
	{
		Data->Checksum = 0;

		DWORD Height = Data->Rect.bottom-Data->Rect.top+1;

		int Bits;
		switch ( Data->Format )
		{
		case D3DFMT_DXT5:
		case D3DFMT_DXT4:
		case D3DFMT_DXT3:
		case D3DFMT_DXT2: 
			Bits = 8; 
			Height /= 4;
			break;
		case D3DFMT_DXT1: 
			Bits = 4;
			Height /= 4;
			break;
		default: 
			{
				char Work[256];
				_snprintf( Work, sizeof(Work), "Unhandled pixel format: %x\n", Data->Format );
				OutputDebugString( Work );
				Bits = 0;
			}
		}

		DWORD Width = ((Data->Rect.right-Data->Rect.left+1) * Bits) / (4*8);
		for ( DWORD y = 0; y < Height; y++ )
		{
			DWORD* Line = (DWORD*)( (DWORD)Data->LockedRect.pBits + y*Data->LockedRect.Pitch );
			for ( DWORD x = 0; x < Width; x++ )
				Data->Checksum += *Line++;
		}

		TextureList.push_back( Data->Checksum );

		map<DWORD,STextureData>::iterator it = TextureBank.find( Data->Checksum );
		if ( it != TextureBank.end() )
		{
			Data->Style = 1; //it->second.Style;
			Data->Color1 = it->second.Color1;
			Data->Color2 = it->second.Color2;
			Data->Rate = it->second.Rate;

			char Work[256];
			_snprintf( Work, sizeof(Work), "Matched texture. Checksum: %x\n", Data->Checksum );
			OutputDebugString( Work );
		}
		else
		{
			Data->Style = 0;
			Save = true;
		}
	}

	HRESULT ret = OldIDirect3DTexture8Vtbl.UnlockRect( _this, Level );

	if ( SaveTextures && Save )
	{
		char Work[256];
		_snprintf( Work, sizeof(Work), "----> Unmatched texture. Checksum: %x <----\n", Data->Checksum );
		OutputDebugString( Work );
		_snprintf( Work, sizeof(Work), "c:\\src\\pshack\\tex\\%x.bmp", Data->Checksum );
		D3DXSaveTextureToFile( Work, D3DXIFF_BMP, (IDirect3DBaseTexture8*)_this, NULL );
	}

	return ret;
}

// IDirect3DDevice8 Interface
IDirect3DDevice8Vtbl HookedIDirect3DDevice8Vtbl;
IDirect3DDevice8Vtbl OldIDirect3DDevice8Vtbl;

IDirect3DTexture8* WhiteTex;

int RenderStyle;

// Used by characters
HRESULT __stdcall IDirect3DDevice8Hook_DrawIndexedPrimitive( 
	IDirect3DDevice8* _this, D3DPRIMITIVETYPE PrimitiveType, UINT minIndex, UINT NumVertices, UINT startIndex, UINT primCount )
{
	HRESULT ret = OldIDirect3DDevice8Vtbl.DrawIndexedPrimitive( _this, PrimitiveType, minIndex, NumVertices, startIndex, primCount );

	RenderStyle %= 3;
	bool Wireframe = RenderStyle != 0;

	for ( int i = 0; i < 4; i++ )
	{
		IDirect3DBaseTexture8* Tex;
		_this->lpVtbl->GetTexture( _this, i, &Tex );

		if ( Tex )
		{
			STextureData* Data = NULL;
			DWORD Size = sizeof(Data);
			HRESULT ok = Tex->lpVtbl->GetPrivateData( Tex, STextureDataGUID, &Data, &Size );
			if ( ok == D3D_OK && Data )
			{
				if ( Data->Style ) Wireframe = true;
				if ( Data->Checksum == *CurTexture ) Wireframe = true;
			}
		}
	}

	if ( Wireframe )
	{
		// Enable wireframe mode.
		DWORD OldFillMode;
		_this->lpVtbl->GetRenderState( _this, D3DRS_FILLMODE, &OldFillMode );
		_this->lpVtbl->SetRenderState( _this, D3DRS_FILLMODE, D3DFILL_WIREFRAME );

		// Use constant white texture.
		IDirect3DBaseTexture8* OldTex;
		_this->lpVtbl->GetTexture( _this, 0, &OldTex );
		_this->lpVtbl->SetTexture( _this, 0, (IDirect3DBaseTexture8*)WhiteTex );

		// Change viewport to force all Z values to the front of the Zbuffer.
		D3DVIEWPORT8 OldViewport;
		D3DVIEWPORT8 Viewport;
		switch ( RenderStyle )
		{
		case 0:
		case 1:
			{
				_this->lpVtbl->GetViewport( _this, &OldViewport );
				Viewport = OldViewport;
				Viewport.MinZ = Viewport.MaxZ = 0.0f; 
				_this->lpVtbl->SetViewport( _this, &Viewport );
				break;
			}
		case 2:
			_this->lpVtbl->SetRenderState( _this, D3DRS_ZBIAS, 1 );
			break;
		}

		OldIDirect3DDevice8Vtbl.DrawIndexedPrimitive( _this, PrimitiveType, minIndex, NumVertices, startIndex, primCount );

		// Restore original states.
		_this->lpVtbl->SetRenderState( _this, D3DRS_ZBIAS, 0 );
		_this->lpVtbl->SetViewport( _this, &OldViewport );
		_this->lpVtbl->SetTexture( _this, 0, OldTex );
		_this->lpVtbl->SetRenderState( _this, D3DRS_FILLMODE, OldFillMode );
	}

	return ret;
}

HRESULT __stdcall IDirect3DDevice8Hook_CreateTexture( 
	IDirect3DDevice8* _this, UINT Width, UINT Height, UINT Levels, 
	DWORD Usage, D3DFORMAT Format, D3DPOOL Pool, IDirect3DTexture8** ppTexture )
{
/*	
	char Work[1024];
	_snprintf( Work, sizeof(Work), "CreateTexture(%d, %d, %d, %d, %x, %d, %x)",
		Width, Height, Levels, Usage, Format, Pool, ppTexture );
	OutputDebugString( Work );
*/

	HRESULT ret = OldIDirect3DDevice8Vtbl.CreateTexture( 
		_this, Width, Height, Levels, Usage, Format, Pool, ppTexture );

	IDirect3DTexture8* Tex = *ppTexture;

	// Allocate space for extra info.
	STextureData* Data = new STextureData;;
	Data->Width = Width;
	Data->Height = Height;
	Data->Format = Format;
	Data->Checksum = 0;
	Tex->lpVtbl->SetPrivateData( Tex, STextureDataGUID, &Data, sizeof(Data), 0 );

	static bool IDirect3DTexture8Hooked = false;
	if ( !IDirect3DTexture8Hooked )
	{
		// Make a copy of the old Vtbl.
		memcpy( &OldIDirect3DTexture8Vtbl, Tex->lpVtbl, sizeof(IDirect3DTexture8Vtbl) );

		// Hook functions.
		DWORD OldProtect;
		VirtualProtect( Tex->lpVtbl, sizeof(IDirect3DTexture8Vtbl), PAGE_READWRITE, &OldProtect );
		Tex->lpVtbl->LockRect = IDirect3DTexture8Hook_LockRect;
		Tex->lpVtbl->UnlockRect = IDirect3DTexture8Hook_UnlockRect;
		VirtualProtect( Tex->lpVtbl, sizeof(IDirect3DTexture8Vtbl), OldProtect, &OldProtect );

		IDirect3DTexture8Hooked = true;
	}

	return ret;
}

bool CycleTexture = false;

void SetCycleMode( bool c )
{
	if ( c != CycleTexture )
	{
		CycleTexture = c;
		char Work[256];
		_snprintf( Work, sizeof(Work), "CycleTexture : %d.\n", CycleTexture );
		OutputDebugString( Work );
	}
}

void OnKey( DWORD Key, DWORD State )
{
	if ( Key == DIK_NUMPADMINUS && (State&0x80) )
	{
		SetCycleMode( false );
		if ( CurTexture == TextureList.begin() )
		{
			OutputDebugString( "CurTexture wrapped." );
			CurTexture = TextureList.end();
		}
		CurTexture--;
	}
	else if ( Key == DIK_NUMPADPLUS && (State&0x80) )
	{
		SetCycleMode( false );
		CurTexture++;
		if ( CurTexture == TextureList.end() )
		{
			OutputDebugString( "CurTexture wrapped." );
			CurTexture = TextureList.begin();
		}
	}
	else if ( Key == DIK_NUMPADPERIOD && (State&0x80) )
	{
		SetCycleMode( !CycleTexture );
	}
	else if ( Key == DIK_NUMPADSTAR && (State&0x80) )
	{
		char Work[256];
		_snprintf( Work, sizeof(Work), "%x\n", *CurTexture );
		OutputDebugString( Work );
	}
	else if ( Key == DIK_F12 && (State&0x80) )
	{
		OutputDebugString( "Toggle RenderStyle.\n" );
		RenderStyle++;
	}
}

HRESULT __stdcall IDirect3DDevice8Hook_Present( 
	IDirect3DDevice8* _this, CONST RECT* pSourceRect, CONST RECT* pDestRect, HWND hDestWindowOverride, CONST RGNDATA* pDirtyRegion )
{
	if ( CycleTexture )
	{
		CurTexture++;
		if ( CurTexture == TextureList.end() )
		{
			OutputDebugString( "CurTexture wrapped." );
			CurTexture = TextureList.begin();
		}
	}

	return OldIDirect3DDevice8Vtbl.Present( _this, pSourceRect, pDestRect, hDestWindowOverride, pDirtyRegion );
}

// IDirect3D8 Interface
IDirect3D8Vtbl HookedIDirect3D8Vtbl;
IDirect3D8Vtbl OldIDirect3D8Vtbl;

HRESULT __stdcall IDirect3D8Hook_CreateDevice( 
	IDirect3D8* _this, UINT Adapter, D3DDEVTYPE DeviceType, HWND hFocusWindow,DWORD BehaviorFlags,
	D3DPRESENT_PARAMETERS* pPresentationParameters, IDirect3DDevice8** ppReturnedDeviceInterface)
{
	pPresentationParameters->FullScreen_PresentationInterval = D3DPRESENT_INTERVAL_ONE;
	HRESULT ret = OldIDirect3D8Vtbl.CreateDevice( 
		_this, Adapter, DeviceType, hFocusWindow, BehaviorFlags, pPresentationParameters, ppReturnedDeviceInterface );

	IDirect3DDevice8* dev = *ppReturnedDeviceInterface;

	static bool IDirect3DDevice8Hooked = false;
	if ( !IDirect3DDevice8Hooked )
	{
		// Make a copy of the old Vtbl.
		memcpy( &OldIDirect3DDevice8Vtbl, dev->lpVtbl, sizeof(IDirect3DDevice8Vtbl) );

		// Create 1x1 white texture.
		D3DLOCKED_RECT LockedRect;
		dev->lpVtbl->CreateTexture( dev, 1, 1, 1, 0, D3DFMT_A8R8G8B8, D3DPOOL_MANAGED, &WhiteTex );
		WhiteTex->lpVtbl->LockRect( WhiteTex, 0, &LockedRect, NULL, 0 );
		*(DWORD*)LockedRect.pBits = 0xFFFF0000;
		WhiteTex->lpVtbl->UnlockRect( WhiteTex, 0 );

		// Hook functions.
		DWORD OldProtect;
		VirtualProtect( dev->lpVtbl, sizeof(IDirect3DDevice8Vtbl), PAGE_READWRITE, &OldProtect );
		dev->lpVtbl->DrawIndexedPrimitive = IDirect3DDevice8Hook_DrawIndexedPrimitive;
		dev->lpVtbl->CreateTexture = IDirect3DDevice8Hook_CreateTexture;
		dev->lpVtbl->Present = IDirect3DDevice8Hook_Present;
		VirtualProtect( dev->lpVtbl, sizeof(IDirect3DDevice8Vtbl), OldProtect, &OldProtect );

		IDirect3DDevice8Hooked = true;
	}

	return ret;
}

// Global functions
typedef IDirect3D8* (WINAPI *Direct3DCreate8_Type)(UINT SDKVersion);
IDirect3D8* WINAPI Direct3DCreate8(UINT SDKVersion);

enum
{
    D3DFN_Direct3DCreate8,
};

// Hook structure.
SDLLHook D3DHook = 
{
    "D3D8.DLL",
    false, 0,
    {
        { "Direct3DCreate8", Direct3DCreate8 },
        { NULL, NULL }
    }
};

IDirect3D8* WINAPI Direct3DCreate8(UINT SDKVersion)
{
    // Poor man's serial port...
    //MessageBeep( MB_ICONHAND );

	// Allow the debugger to connect.
	if ( GetAsyncKeyState( VK_SHIFT ) )
	{
		MessageBox( NULL, "Attach debugger to target process, and press OK.", "PSHack", MB_ICONHAND | MB_OK );
		DebugBreak();
	}

	ReadINIFile();

    Direct3DCreate8_Type OldFn = 
        (Direct3DCreate8_Type)D3DHook.Functions[D3DFN_Direct3DCreate8].OrigFn;
    IDirect3D8* ret = OldFn(SDKVersion);

/*
    Other way to do it:

    // Make copies of the old Vtbl.
    memcpy( &OldIDirectDrawVtbl, ( *lplpDD )->lpVtbl, sizeof(IDirectDrawVtbl) );
    memcpy( &HookedIDirectDrawVtbl, ( *lplpDD )->lpVtbl, sizeof(IDirectDrawVtbl) );

    // Create the hooked vtbl.
    HookedIDirectDrawVtbl.QueryInterface = IDirectDrawHook_QueryInterface;

    // Reroute the Vtbl.
    ( *lplpDD )->lpVtbl = &HookedIDirectDrawVtbl;
*/

    static bool IDirect3D8Hooked = false;
    if ( !IDirect3D8Hooked )
    {
        // Make a copy of the old Vtbl.
        memcpy( &OldIDirect3D8Vtbl, ret->lpVtbl, sizeof(IDirect3D8Vtbl) );

        // Hook functions.
		DWORD OldProtect;
		VirtualProtect( ret->lpVtbl, sizeof(IDirect3D8Vtbl), PAGE_READWRITE, &OldProtect );
        ret->lpVtbl->CreateDevice = IDirect3D8Hook_CreateDevice;
		VirtualProtect( ret->lpVtbl, sizeof(IDirect3D8Vtbl), OldProtect, &OldProtect );

        IDirect3D8Hooked = true;
    }

    return ret;
}
